# Copyright (c) Ran Dugal 2014
#
# This file is part of dust.
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

'''
Dust command loop
'''

import sys
import socket
import os
import readline
import logging
import stat
import colorama
import threading

from collections import defaultdict
from cmd import Cmd
from EC2 import EC2Config

import paramiko
from dustcluster import commands, lineterm
from dustcluster import __version__
from dustcluster.cluster import ClusterCommandEngine
from dustcluster.config import DustConfig

import atexit
import shlex

from dustcluster import util
logger = util.setup_logger( __name__ )

if os.environ.get('COLORTERM') or 'color' in os.environ.get('TERM'):
    colorama.Fore.CYAN  = '\x1b[38;5;75m'
    colorama.Fore.GREEN = '\x1b[38;5;76m'

class Console(Cmd):
    ''' command line tool to control a cloud cluster '''

    dustintro  = "Dust cluster shell, version %s. Type %s?%s for help." % (__version__, colorama.Fore.GREEN, colorama.Style.RESET_ALL)

    def __init__(self):

        util.intro()

        logger.setLevel(logging.INFO)

        # read/create config
        try:
            self.config = DustConfig()
        except Exception, e:
            logger.error("Error getting config/credentials. Cannot continue.")
            raise

        # load history
        try:
            if os.path.exists(self.config.get_history_file_path()):
                readline.read_history_file(self.config.get_history_file_path())

        except IOError:
            logger.warning("Error reading history file. No command history available.")

        atexit.register(self.on_exit, None)

        Cmd.__init__(self)

        self.commands = {}  # { cmd : (helpstr, module) }
        # startup
        self.cluster = ClusterCommandEngine()
 
        self.exit_flag = False
        self.cluster.lineterm.set_refresh_callback(self.redisplay)

        threading.Thread(target=self.cluster.load_commands).start()
        logger.info(self.dustintro)

    @property
    def prompt(self):
        if self.cluster.cloud and self.cluster.cloud.region:
            return "[%s]$ " % self.cluster.cloud.region
        else:
            return "[dust]$ "


    def redisplay(self):
        ''' refresh the prompt '''
        sys.stdout.write('\n\r' + self.prompt)
        sys.stdout.write(readline.get_line_buffer())
        sys.stdout.flush()

    def on_exit(self, _ ):
        readline.write_history_file(self.config.get_history_file_path())

        confregion = self.config.user_data['region']
        cloud_region = ""
        if self.cluster and self.cluster.cloud:
            cloud_region = self.cluster.cloud.region

        if confregion != cloud_region:
            self.config.user_data['region'] = cloud_region 
            self.config.write_user_data()
            
        print
        
    def emptyline(self):
        pass

    def do_help(self, args):
        '''
        Modified from base class.
        '''
        self.default("help " + args)

    def default(self, line):

        cmd, arg, line = self.parseline(line)
        ret = self.cluster.handle_command(cmd, arg)

        if not ret:
            # not handled? try system shell
            logger.info( 'dustcluster: [%s] unrecognized, trying system shell...\n' % line )
            os.system(line)

        return


    def parseline(self, line):

        # expand @target cmd to atssh target cmd
        if line and line[0] == '@':
            tokens = line.split()
            if len(tokens[0]) == 1:
                line = 'atssh * ' + line[1:]
            else:
                target  = tokens[0][1:]
                line = 'atssh %s %s ' % (target, line[len(tokens[0]):] )

        ret = Cmd.parseline(self, line)
        return ret


    def do_exit(self, _):
        '''
        exit - exit dust shell
        '''     
        logger.info( 'Exiting dust console. Find updates, file bugs at http://github.com/carlsborg/dust.')
        logger.info( '%sThis is an early beta release. Consider updating with $pip install dustcluster --upgrade"%s.' %
                     (colorama.Fore.GREEN, colorama.Style.RESET_ALL))
        
        self.cluster.logout()
        self.exit_flag = True
        return True

    def do_EOF(self, line):
        '''
        EOF/Ctrl D - exit dust shell
        '''
        return self.do_exit(line)




# Copyright (c) Ran Dugal 2014
#
# This file is part of dust.
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import colorama

''' dust commands to start/stop/terminate nodes '''

# export commands
commands  = ['show', 'refresh', 'start', 'stop', 'terminate']


def show(cmdline, cluster, logger):
    '''
    show  [-vv] [filter | search_term]  - show all nodes or filtered nodes

    List node data from in memory cache. If filter does not match a node index,
    cluster name, node name, or EC2 attribute key=value expression then search
    all attributes and tags.

    Use $refresh to update cache.  
    Shows only nodes selected by the use command, if it was invoked earlier.

    -v  show more attributes
    -vv show all attributes

    Example:    
    show 1,3,4                   # filter_exp matches index numbers
    show worker[3-15]            # filter_exp matchs node names
    show launch_time=2017-10*    # filter_exp matches EC2 attribute 
    show 192.168.0.1             # no filter_exp match, use as search term 
    show state=running           # filter_exp matches 
    show running                 # search term matches

    '''

    args = cmdline.split()
    extended = 0
    if args and args[0].startswith("-"):
        if args[0] == "-v":
            extended = 1
        elif args[0] == "-vv":
            extended = 2
        else:
            logger.error("Unknown switch %s" % args[0])
            return
 
        cmdline = " ".join(args[1:])

    if not cmdline.strip():
        cmdline = '*'

    _show(cmdline, cluster, logger, extended)

def _show(cmdline, cluster, logger, extended=0):
    try:

        target_nodes = []
        target_nodes = get_target_nodes(logger, cluster, cmdline, search=True)
        if not target_nodes:
            return

        cluster.show(target_nodes, extended=extended)

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

def refresh(cmdline, cluster, logger):
    '''
    refresh [filter]  - refresh from cloud and call show with filter

    Note that some operations (start/stop/etc) cause a refresh to occur on the next show.
    '''

    cluster.invalidate_cache()
    _show(cmdline, cluster, logger, False)

def start(cmdline, cluster, logger):
    '''   
    start filter   - start nodes or restart stopped nodes

    Arguments:
    none    --- Start all nodes defined in the cluster, idempotently
    filter  --- A node name or filter expression (see help filters) 
                Node names and filter values can be regular expressions.

    Example:
    start 3,4,5
    start devdb,dev1,dev2
    start state=stopped
    start worker[0-10]
    start cluster=spark1
    '''

    if not cmdline.strip():
        logger.error("start filter  - Restart stopped nodes")
        return

    try:

        target_nodes = get_target_nodes(logger, cluster, cmdline)

        if not target_nodes:
            logger.info('no target nodes to operate on')
            return

        for node in target_nodes:

            if not node.hydrated and not node.key:
                logger.info("No key name configured for this node in the template. Need a key name to launch a node.")
                keyname = raw_input("Keyname [Enter to use dustcluster default]:")
                if keyname:
                    node.key = keyname
                else:
                    node.key, keyfile = cluster.get_default_key()
                    logger.info("Using default key [%s] in [%s]" % (node.key, keyfile)) 

            node.start()

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

    logger.info( 'ok' )

    cluster.invalidate_cache()

def stop(cmdline, cluster, logger):
    '''
    stop  filter   - stop nodes

    Arguments:
    none    --- Stop all nodes defined in the cluster, idempotently
    filter  --- A node name or filter expression 

    Example:
    stop failover1
    stop state=running
    stop worker[0-20]
    stop worker*
    '''

    if not cmdline.strip():
        logger.error("stop filter   - Stop nodes")
        return
    
    operation(logger, cluster, 'stop', cmdline, confirm=True)
    cluster.invalidate_cache()

def terminate(cmdline, cluster, logger):
    '''
    terminate filter  - terminate nodes 

    Arguments:
    none    --- Terminate all nodes defined in the cluster, idempotently
    filter  --- A node name or filter expression (see help filters) 
                Node names and filter values can be regular expressions.

    Example:
    terminate failover1
    terminate state=running
    terminate worker[0-20]
    terminate worker*
    '''

    if not cmdline.strip():
        logger.error("terminate filter  - Terminate nodes")
        return

    operation(logger, cluster, 'terminate', cmdline, confirm=True)
    cluster.invalidate_cache()

def operation(logger, cluster, op, target_node_str=None, confirm=False):
    ''' invoke attribute op on a set of nodes ''' 

    try:

        target_nodes = get_target_nodes(logger, cluster, target_node_str)

        if not target_nodes:
            logger.info('no target nodes to operate on')
            return

        if confirm:
            cluster.show(target_nodes)
            logger.info( "%sInvoking %s on these nodes%s" % (colorama.Fore.RED, op, colorama.Style.RESET_ALL) )
            confirm_str = "Continue [Y/N]:"
            s = raw_input(confirm_str) 
            while( s.lower() != 'y' and s.lower() != 'n' ):
                s = raw_input(confirm_str) 

            if s.lower() == 'n':
                return

        for node in target_nodes:
            if node.state != "terminated":
                getattr(node, op)()

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

    logger.info( 'ok' )


def get_target_nodes(logger, cluster, target_node_str=None, search=False):

    target_nodes = cluster.resolve_target_nodes(search=search, target_node_name=target_node_str)

    return target_nodes




# Copyright (c) Ran Dugal 2014
#
# This file is part of dust cluster
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import logging
import json
import yaml
from collections import namedtuple
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager
from ansible.inventory.manager import InventoryManager
from ansible.playbook.play import Play
from ansible.executor.task_queue_manager import TaskQueueManager
from ansible.plugins.callback import CallbackBase

''' install spark on a base cluster '''

# export commands 
commands  = ['install_spark']


class ResultCallback(CallbackBase):
    def v2_runner_on_ok(self, result, **kwargs):
        """Print a json representation of the result

        This method could store the result in an instance attribute for retrieval later
        """
        host = result._host
        print(json.dumps({host.name: result._result}, indent=4))

Options = namedtuple('Options', ['connection', 'module_path', 'forks', 'become', 'become_method', 'become_user', 'check', 'diff'])


def install_spark(cmdline, cluster, logger):
    '''
    install_spark master workers - install spark on nodes master, workers 
    '''

    args = cmdline.split()
    if len(args) < 2:
        print "usage: install_spark master_node_filter worker_node_filter"
        return

    # initialize needed objects
    loader = DataLoader()
    options = Options(connection='ssh', module_path="", forks=100, become=None, become_method=None, become_user=None, check=False, diff=False)

    # Instantiate our ResultCallback for handling results as they come in
    results_callback = ResultCallback()

    # create inventory and pass to var manager
    inventory = InventoryManager(loader=loader, sources=['/home/booda/hosts'])
    variable_manager = VariableManager(loader=loader, inventory=inventory)

    # create play with tasks
    with open('/home/booda/dev/dust/p2.yaml') as data_file:    
        data = yaml.load(data_file)

    play_source = data
    play = Play().load(play_source, variable_manager=variable_manager, loader=loader)

    # actually run it
    tqm = None
    try:
        tqm = TaskQueueManager(
                inventory=inventory,
                variable_manager=variable_manager,
                loader=loader,
                options=options,
                passwords = None, 
                stdout_callback=None 
            )

        result = tqm.run(play)
    finally:
        if tqm is not None:
            tqm.cleanup()



import subprocess
import sys
import os
import threading
import time
import logging


class ShellExecutorThread(threading.Thread):

    def __init__(self, read_file, cmd):
        self.readfrom = read_file
        self.proc = None
        self.cmd = cmd
        super(ShellExecutorThread, self).__init__()

    def run(self):
        self.proc = subprocess.Popen(self.cmd, stdin=self.readfrom, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True, shell=True)
        self.proc.wait()


class ShellPipeOperation(object):

    def __init__(self):

        r, w = os.pipe()
        self.pipe_write_file = os.fdopen(w, 'w')
        self.pipe_read_file  = os.fdopen(r, 'r')
        self.old_stdout = None
        self.old_stderr = None
        self.tgt = None

    def invoke_shell_cmd(self, cmd):
        ''' 
            invokes a shell command in its own thread and pipes stdout to it 
            returns when proc is spawned
        '''

        self.tgt = ShellExecutorThread(self.pipe_read_file, cmd)
        self.tgt.start()

        while not self.tgt.proc:
            time.sleep(0.2)

    def redirect_stdout_to_shell(self):
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        sys.stdout = self.pipe_write_file
        sys.stderr = self.pipe_write_file

    def reset_stdout(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    def get_shell_output(self):
        out, err = self.tgt.proc.communicate()
        return out, err

    def cleanup(self):
        self.pipe_write_file.close()
        self.tgt.join()

